from .borders import Bdr
from .context import Context
from .text import Text
from .line import Line, Just, justify_v
from .widget import INFINITY

class HR(Text, border=Bdr.dfn(l='\n\n│\n┤\n\n[\n[', r='\n\n│\n├\n\n]\n]', t='─-')):

    def calc_width(self):
        w = self['width_max']
        if not w or w == INFINITY:
            w = super().calc_width(Text) + 6
            if w == 6: return 3
        return w


    def _flatten_empty(self, w, h, ascii, bdr, style):
        w = w or self['width_max']
        if not w or w == INFINITY: w = 3
        bar = bdr.t[7 if ascii else 2]
        bar_l = bdr.t[5 if ascii else 0]
        bar_r = bdr.t[8 if ascii else 3]
        line = bar if w==1 else (bar_l+bar_r) if w==2 else (bar_l + bar*(w-2) + bar_r)
        yield from justify_v([Line(parent=self)(line,style=style)], h, Just(self['justify'], '~'), Line(parent=self).insert(0,' '*w))
        

    def flatten(self, w=0, h=0, **kwargs):
        flat = list(super().flatten(w=0, h=h, **kwargs))
        ascii, bdr, style = self['ascii'], self['border'], self['border_style']
        if not any(flat):
            yield from self._flatten_empty(w,h, ascii, bdr, style)
            return
        mw = w or self['width_max']
        if mw == INFINITY: mw = 0
        if mw and mw < 7:
            yield from super().flatten(w=mw, h=h, **kwargs)
            return
        mw = mw or (flat[0].width + 6)
        if mw-6 < flat[0].width:
            flat = list(super().flatten(w=mw-6, h=h, **kwargs))
        innerw = flat[0].width
        just = Just(self['just'],'<~')
        hrv = just.pad_v(len(flat)-1)
        bar = bdr.t[7 if ascii else 2]
        bar_l = bdr.t[5 if ascii else 0]
        bar_r = bdr.t[8 if ascii else 3]
        box_l = bdr.l[7 if ascii else 2]
        join_l = Bdr.join(Bdr.ext(box_l+' '+box_l+bar))
        if join_l == ' ': join_l = box_l
        box_r = bdr.r[7 if ascii else 2]
        join_r = Bdr.join(Bdr.ext(box_r+bar+box_r+' '))
        if join_r == ' ': join_r = box_r
        left = just.pad_h(mw-innerw) or 3
        if left == mw-innerw: left = mw-innerw-3
        for i,l in enumerate(flat):
            if i == hrv:
                lhs = bar_l + bar*(left-3) + join_l + ' '
                rhs = ' ' + join_r + bar*(mw-innerw-left-3) + bar_r
            else:
                lhs = ' '*(left-2) + box_l + ' '
                rhs = ' ' + box_r + ' '*(mw-innerw-left-2)
            yield Line(parent=self)(lhs,style=style)(l)(rhs,style=style)




import pytest, io
from functools import reduce
from print_ext.table import Table, CellDfn
from print_ext.fill import Fill
from print_ext.context import Context
from print_ext.printer import Printer
from print_ext.line import SMark as SM
from print_ext.borders import Bdr
from .testutil import debug_dump
from .printer_test import _printer

def test_table_x():
    t = Table(3, -1, 5, tmpl='', ascii=True, wrap=False)
    t('test\tx\t')
    t(33,'\t',Fill('a\vb'),'\t',Fill('.'), '\t', 'tは quick brown fox jumped over the lazy dog')
    for line, expect in zip(t.flatten(w=10), ['t~tx33    ', 'aaa.tは~og', 'bbb.      ']):
        assert(line.styled()[0] == expect)
        


def test_table_x2():
    t = Table(3, -1, 5, ascii=True, tmpl='', wrap=True)
    t('test\tx\t')
    t(33,'\t',Fill('a\vb'),'\t',Fill('.'), '\t', 'The quick brown fox jumped over the lazy dog')
    for line, expect in zip(t.flatten(w=15), ['tesx33         ', 't              ', 'aaa.The quick b', 'bbb.\\ rown fox ','aaa.\\ jumped ov','bbb.\\ er the la','aaa.\\ zy dog   ']):
        assert(line.styled()[0] == expect)


def test_table_clone():
    t = Table(0.0, 1, tmpl='')
    t('a\tb\tc\td\t')
    b = t.clone(**t.ctx_local)
    assert([r.styled()[0] for r in b.flatten()] == ['ab','cd'])


def test_table_last_tab():
    t = Table(1,1)
    t('0\t1\t2\t3\t4\t')
    f = list(t.flatten())
    assert(len(f) == 3)
    t = Table(1,1)
    t('0\t1\t2\t3\t')
    f = list(t.flatten())
    assert(len(f) == 2)
    

def test_table_templates():
    t = Table(1,1)
    t('1\t2\ta\tb\t')
    flat = [f.styled() for f in t.flatten()]
    print('\n'.join(f'-{x}-' for x in flat))
    assert(flat[0] == ('1 2',[SM('em',0,3), SM('dem',1,2)]))
    assert(flat[1] == ('a b',[SM('dem',1,2)]))



def test_table_styles():    
    t = Table(-1, -2, -3, ascii=True, tmpl='', style='r', wrap=False)
    t.cell('R0', style='1')
    t.cell('C1', style='y')
    t.cell('R1%2C-1', style='_')
    t('A0\tA1\tA2\t')
    t('B0\tB1\tB2\t')
    t('C0\tC1\tC2\t')
    t('D0\tD1\tD2\t')
    rows = [r.styled() for r in t.flatten()]
    assert(rows[0] == ('~A1A2 ',[SM('r',0,6), SM('1',0,6), SM('dem',0,1), SM('y',1,3)]))
    assert(rows[1] == ('~B1B2 ', [SM('r',0,6), SM('dem',0,1), SM('y',1,3), SM('_',3,6)]))
    assert(rows[2] == ('~C1C2 ', [SM('r',0,6), SM('dem',0,1), SM('y',1,3)]))
    assert(rows[3] == ('~D1D2 ', [SM('r',0,6), SM('dem',0,1), SM('y',1,3), SM('_',3,6)]))
    assert(len(t) == 12)
    


def test_table_play():  
    o,p = _printer(color=True)  
    t = Table(-1, -2, -3, tmpl='pad', ascii=True, wrap=False)
    print(t['tmpl'])
    t('A0\tA1\tA2\tB0\t\b^r$ B1\tB2\t', 'C0\t\b$ C1\tC2\t')
    t('D0\tD1\tD2\t')
    assert([s.styled()[0] for s in t.flatten()] == ['~ ~ A2', '~ ~ B2', '~ ~ C2', '~ ~ D2'])
    assert([s.styled()[0] for s in t.flatten()] == ['~ ~ A2', '~ ~ B2', '~ ~ C2', '~ ~ D2'])
    p(t)
    print(o.getvalue())



def test_table_CellDfn_ctx_extend():
    cells = [ CellDfn('R0', cls=Fill, style='1-2'), CellDfn('C1', cls=Table, style='3'), CellDfn('R0C2', style='4')]
    ctx = Context()
    assert(reduce(lambda a, c: c.ctx_merge(a, 4, 4, 10, 10), cells, ctx) == ctx)
    assert(reduce(lambda a, c: c.ctx_merge(a, 0, 4, 10, 10), cells, ctx)['cls'] == Fill)
    assert(reduce(lambda a, c: c.ctx_merge(a, 0, 1, 10, 10), cells, ctx)['cls'] == Table)
    assert(reduce(lambda a, c: c.ctx_merge(a, 0, 2, 10, 10), cells, ctx)['style'] == ('1','2','4'))



def test_table_CellDfn():
    for s in ['r0', 'R-1', 'R%3', 'R3%4', 'C3', 'R0C0', 'all']:
        print(CellDfn(s))
    for s in ['0', 'C0R0']:
        with pytest.raises(ValueError):
            CellDfn(s)

    c = CellDfn('R0')
    for rc in [(0,0), (0,1), (0, 99)]:
        assert(c.matches(*rc, 10,10))
    for rc in [(1,0), (3,1), (5, 99)]:
        assert(not c.matches(*rc, 10,10))
    
    c = CellDfn('R-4%2C3%3')
    for rc in [(6,3), (8,6), (6, 9)]:
        assert(c.matches(*rc, 10,10))
    for rc in [(0,0), (8,0)]:
        assert(not c.matches(*rc, 10,10))



def test_sudoku():
    o,p = _printer()
    b = Table(1,1,1,1,1,1,1,1,1, tmpl='grid,dbl')
    b.cell('C3%3', border=('l:║#','c:╫\n\n\n#\n\n\n'))
    b.cell('R3%3', border=('t:═#','c:╪\n\n\n#\n\n\n'))
    b.cell('R3%3C3%3', border=('c:╬\n\n\n#\n\n\n'))
    b.cell('R0C3%3', border=('c:╦\n\n\n#\n\n\n'))
    b.cell('R3%3C0', border=('c:╠\n\n\n#\n\n\n'))
    b.cell('R-1C3%3', border=('c:\n\n╩\n\n\n#\n'))
    b('１\t２\t３\t４\t５\t６\t７\t８\t９\t')
    b('１\t２\t３\t４\t５\t６\t７\t８\t９\t')
    b('１\t２\t３\t４\t５\t６\t７\t８\t９\t')
    b('１\t２\t３\t４\t５\t６\t７\t８\t９\t')
    b('１\t２\t３\t４\t５\t６\t７\t８\t９\t')
    b('１\t２\t３\t４\t５\t６\t７\t８\t９\t')
    b('１\t２\t３\t４\t５\t６\t７\t８\t９\t')
    b('１\t２\t３\t４\t５\t６\t７\t８\t９\t')
    b('１\t２\t３\t４\t５\t６\t７\t８\t９\t')
    p(b)
    print(o.getvalue())
    assert(o.getvalue() == '╔══╤══╤══╦══╤══╤══╦══╤══╤══╗\n║１│２│３║４│５│６║７│８│９║\n╟──┼──┼──╫──┼──┼──╫──┼──┼──╢\n║１│２│３║４│５│６║７│８│９║\n╟──┼──┼──╫──┼──┼──╫──┼──┼──╢\n║１│２│３║４│５│６║７│８│９║\n╠══╪══╪══╬══╪══╪══╬══╪══╪══╢\n║１│２│３║４│５│６║７│８│９║\n╟──┼──┼──╫──┼──┼──╫──┼──┼──╢\n║１│２│３║４│５│６║７│８│９║\n╟──┼──┼──╫──┼──┼──╫──┼──┼──╢\n║１│２│３║４│５│６║７│８│９║\n╠══╪══╪══╬══╪══╪══╬══╪══╪══╢\n║１│２│３║４│５│６║７│８│９║\n╟──┼──┼──╫──┼──┼──╫──┼──┼──╢\n║１│２│３║４│５│６║７│８│９║\n╟──┼──┼──╫──┼──┼──╫──┼──┼──╢\n║１│２│３║４│５│６║７│８│９║\n╚══╧══╧══╩══╧══╧══╩══╧══╧══╝\n')


@pytest.mark.skip(reason="not implemented")
def test_table_partial_show():
    ''' Show a portion of the table while building.  The col widths may change, so use fixed col widths
    '''
    print.card('\berr Error:\t', *args, **kwargs)


def test_table_coldfns():
    t = Table(-3,-4,tmpl='',ascii=True,wrap=False)
    t('12345\t12345\t')
    f = [s.styled()[0] for s in t.flatten()][0]
    assert(f == '1~512~5')



import pytest
from print_ext.text import Text
from print_ext.line import Line, SMark as SM
from .testutil import debug_dump, styled, flat



def test_text_nl_style():
    ''' when adding a blank newline the style should still be applied '''
    t=Text('\v\v\v', style='x')
    for row in t.flatten(w=4):
        assert(row.styled() == ('    ', [SM('x',0,4)]))
        


def test_text_clone():
    p = Text()
    t = Text('a\vb')
    y = t.clone(parent=p)
    assert(y != t)
    assert(y.parent == p)
    m = Text(t, t)
    assert(str(m) == 'a\\nba\\nb')



@pytest.mark.skip(reason="Disabled auto-unindent")
def test_text_nl():
    def _f(*args, w=0, h=0, **kwargs): return ','.join(map(str, Text(*args, **kwargs).flatten(w=w,h=h)))
    assert(_f('''  The
        quick
        brown
          fox
    ''', justify='>') == '  The,quick,brown,  fox')
    assert(_f('\na\nb\n') == 'a,b')
    assert(_f('   a\nb ') == 'a ,b ')



def test_width_height():
    t = Text('The\vquick brown\vfox\v')
    assert(t.width == 11)
    assert(t.height == 4)
    t = Text()
    assert(t.width == 0)
    assert(t.height == 0)
    t = Text('bob')
    assert(t.width == 3)
    assert(t.height == 1)
    t = Text('\v\v')
    assert(t.width == 0)
    assert(t.height == 3)



def test_text_flatten():
    t = Text('the', ' quick\v', 'brown', text_wrap=True, wrap_mark_from=10)
    assert('--'.join(flat(t, w=5)) == 'the q--uick --brown')
    assert(flat(Text()) == [])
    assert(flat(Text('', '', '')) == [])



def test_text_styled():
    t = Text('\bx$ hi', '\vbye', '\v\by bob', 'coj', style='t', wrap=False, ascii='on')
    print('\n'.join(debug_dump(t)))
    assert(styled(t, w=3, h=4) == [
        ('hi ', [SM('t',0,3), SM('x',0,3)]),
        ('bye', [SM('t',0,3), SM('x',0,3)]),
        ('b~j', [SM('t',0,3), SM('x',0,1), SM('y',0,1), SM('dem',1,2), SM('x',2,3)]),
        ('   ', [SM('t',0,3)]),
    ])



def test_text_e():
    p = Text('parent')
    t= Text('ERROR', parent=p, style='err')
    #print('\n'.join(debug_dump(t)))
    assert(styled(t) == [('ERROR', [SM('err',0,5)])])



def test_empty_flattenable():
    s = Text()
    t = Text(s)
    assert(str(t) == '')


@pytest.mark.skip(reason='Not Implemented')
def test_text_lang():
    s = Text('\berr$', 'Danger\fzh 危险', '\b$ !', "Don't hold plutonium\vwith bare hands.\fzh 不要赤手拿钚。")
    assert(flat(s) == ["Danger!Don't hold plutonium", 'with bare hands.           '])
    s = Text('\berr$', 'Danger\fzh 危险', '\b$ !', "Don't hold plutonium\vwith bare hands.\fzh 不要赤\v手拿钚。", lang='zh')
    assert(flat(s) == ['危险!不要赤','手拿钚。   '])



from functools import reduce
from .context import Context, CVar
from .widget import Widget


class ProxyContext(Context):
    def __init__(self, eoc=True, **kwargs):
        self.eoc = eoc
        super().__init__(**kwargs)

    def is_empty(self):
        return not self.ctx_local and (not self.parent or self.parent.is_empty())

    def ctx_flatten(self):
        return {} if self.is_empty() else super().ctx_flatten()

        

class LangCVar(CVar):
    def canon(self, val):
        if isinstance(val, str):
            return tuple((val.lower()+'___').split('_')[:3])
        return val

Context.define(LangCVar('lang'))



class Rich(Widget):
    ''' This is a mixin for adding rich-text parsing capabilities via the __call__ method.

    :Special characters:
        \\b
            Put an attribute on the stack.
            `\\b<key>=<value><auto-pop>`

            `\\b=R`   This gets auto-popped at the end of the call
            `\\b=y$`  This gets auto-popped at the end of the string
            `\\b=r!`  This doesn't get auto-popped.  Pop it manually with 
            `\\b=`    Pop an attribute.  It is ok to pop too many times.

        \\t
            Triggers a move to the next 'cell'.  The meaning of 'cell' is defined by the baseclass.

            
    :Usage:
        Subclass this and then implement the following methods as a state machine.
            rich_append(el)
                el will be flattenable
            rich_newline()
                This is optional.  By default throws.
            rich_tab()
                This is optional.  By default throws.
    '''

    def __init__(self, *args, **kwargs):
        self.rich_stream = []
        super().__init__(**kwargs)
        self(*args)


    def __call__(self, *els, **kwargs):
        return self.append(*els, **kwargs)


    def append(self, *els, **kwargs):
        ''' Add any kind of element to this.

        If el is a str then it is parsed as rich-text. (b,f,t,v escape codes)
        If it is a Context object, then it is added as is.
        Otherwise, turn it into a Line() and add that.
        '''
        self.ctx_cur = ProxyContext(**kwargs)
        for el in els:
            if isinstance(el, str):
                lines = el.replace('\v','\n').split('\n')
                #lines = _unindent(el).replace('\v','\n').split('\n') # FIXME: Removed self.__l10n(el) and unindent
                self.__append_line(lines[0])
                for line in lines[1:]:
                    self.rich_newline(self.ctx_cur.ctx_flatten())
                    self.__append_line(line)
                while self.__pop(False): pass # Pop end-of-string
            elif isinstance(el, Context):
                self.rich_append(el, self.ctx_cur.ctx_flatten())
            else:
                self.rich_append(str(el), self.ctx_cur.ctx_flatten())
        self.changed_size()
        return self


    def __append_line(self, line):
        tabs = line.split('\t')
        self.__append_tab(tabs[0])
        for tab in tabs[1:]:
            self.rich_tab(self.ctx_cur.ctx_flatten())
            self.__append_tab(tab)
    

    def __append_tab(self, tab):
        parts = tab.split('\b')
        if parts[0]: self.rich_append(parts[0], self.ctx_cur.ctx_flatten())
        for part in parts[1:]:
            part = part.split(' ',1)
            eoc = part[0].endswith('$')
            part, txt = part[0][:-1] if eoc else part[0], (part[1] if len(part) > 1 else '')
            if part:
                part = part.split('=')
                kwargs = {'':part[0]} if len(part) == 1 else {part[0]:part[1]}
                self.ctx_cur = ProxyContext(eoc=eoc, parent=self.ctx_cur, **kwargs)
            else: # empty \b
                self.__pop(eoc)
            if txt: self.rich_append(txt, self.ctx_cur.ctx_flatten())


    def __pop(self, eoc):
        pctx, ctx = None, self.ctx_cur
        while ctx.parent and ctx.eoc != eoc:
            pctx, ctx = ctx, ctx.parent
        if not ctx.parent: return False
        if pctx:
            pctx.parent = ctx.parent
        else:
            self.ctx_cur = ctx.parent
        return True


    def __l10n(self, txt):
        vers = txt.split('\f')
        best = (vers[0], 0.5)
        glang = self['lang'] or ('','','')
        for ver in vers[1:]:
            try:
                lang, ver = ver.split(' ',1)
            except Exception as e:
                raise ValueError(f"Invalid \\f usage '{ver}'\nExample: '\\fen_us color\\fen_gb colour'")
            slang = (lang.lower()+'___').split('_')[:3]
            
            m = 0 if slang[0] != glang[0] else\
                1 if slang[1] != glang[1] else\
                2 if slang[2] != glang[2] else\
                3
            m += reduce(lambda a,x: a + (0.0 if x else 0.1), slang, 0)
            if m > best[1]: best = (ver, m)
        return best[0]


    def rich_append(self, el, ctx):
        try:
            el = el.ctx_parent(self)
        except AttributeError:
            pass
        self.rich_stream.append((el,ctx))
        

    def rich_newline(self, ctx):
        self.rich_stream.append(('\n',ctx))
    

    def rich_tab(self, ctx):
        self.rich_stream.append(('\t',ctx))

    

def _unindent(txt):
    if '\n' not in txt: return txt
    newlines = txt.splitlines()
    indent = ''
    for line in newlines[1:]: # The first line is skipped because it starts right after the opening '''
        stripped = line.lstrip()
        if stripped: # First non-blank line sets the indentation level
            indent = line[:len(line)-len(stripped)]
            break
    n_indent = len(indent)
    if indent: newlines = [l[n_indent:] if l.startswith(indent) else l for l in newlines]
    while newlines and not newlines[-1].strip(): newlines = newlines[:-1]
    return '\n'.join(newlines).lstrip()



from .printer_test import _printer
from print_ext.printer import Printer
from print_ext.table import Table
import cProfile, re, pstats, sys

print = Printer(color=True, width=500)
hard = {'apiVersion': 'networking.k8s.io/v1', 'kind': 'Ingress', 'metadata': {'name': 'toss-ingress', 'namespace': 'toss', 'annotations': {'kubernetes.io/ingress.global-static-ip-name': 'toss-ip', 'nginx.ingress.kubernetes.io/proxy-read-timeout': '120'}}, 'spec': {'tls': [{'secretName': 'toss-online-com-certbot'}], 'rules': [{'host': 'm-test.tos-land.net', 'http': {'paths': [{'path': '/', 'pathType': 'Prefix', 'backend': {'service': {'name': 'prod-api-service', 'port': {'number': 9000}}}}]}}, {'host': 'tos-land.net', 'http': {'paths': [{'path': '/', 'pathType': 'Prefix', 'backend': {'service': {'name': 'prod-api-service', 'port': {'number': 9000}}}}]}}, {'host': 'www.tos-land.net', 'http': {'paths': [{'path': '/', 'pathType': 'Prefix', 'backend': {'service': {'name': 'prod-api-service', 'port': {'number': 9000}}}}]}}, {'host': 'm.tos-land.net', 'http': {'paths': [{'path': '/', 'pathType': 'Prefix', 'backend': {'service': {'name': 'prod-api-service', 'port': {'number': 9000}}}}]}}, {'host': 'test-video.toss-online.com', 'http': {'paths': [{'path': '/', 'pathType': 'Prefix', 'backend': {'service': {'name': 'test-frontend-service', 'port': {'number': 9104}}}}, {'path': '/vkey', 'pathType': 'Prefix', 'backend': {'service': {'name': 'test-api-service', 'port': {'number': 9100}}}}, {'path': '/api', 'pathType': 'Exact', 'backend': {'service': {'name': 'test-api-service', 'port': {'number': 9100}}}}, {'path': '/apitest', 'pathType': 'Exact', 'backend': {'service': {'name': 'test-api-service', 'port': {'number': 9100}}}}, {'path': '/download', 'pathType': 'Prefix', 'backend': {'service': {'name': 'test-api-service', 'port': {'number': 9100}}}}, {'path': '/vinfo', 'pathType': 'Prefix', 'backend': {'service': {'name': 'test-api-service', 'port': {'number': 9100}}}}]}}, {'host': 'video.toss-online.com', 'http': {'paths': [{'path': '/', 'pathType': 'Prefix', 'backend': {'service': {'name': 'prod-frontend-service', 'port': {'number': 9004}}}}, {'path': '/vkey', 'pathType': 'Prefix', 'backend': {'service': {'name': 'prod-api-service', 'port': {'number': 9000}}}}, {'path': '/api', 'pathType': 'Exact', 'backend': {'service': {'name': 'prod-api-service', 'port': {'number': 9000}}}}, {'path': '/apitest', 'pathType': 'Exact', 'backend': {'service': {'name': 'test-api-service', 'port': {'number': 9100}}}}, {'path': '/download', 'pathType': 'Prefix', 'backend': {'service': {'name': 'prod-api-service', 'port': {'number': 9000}}}}, {'path': '/vinfo', 'pathType': 'Prefix', 'backend': {'service': {'name': 'prod-api-service', 'port': {'number': 9000}}}}]}}, {'host': 'api.toss-online.com', 'http': {'paths': [{'path': '/apitest', 'pathType': 'Exact', 'backend': {'service': {'name': 'test-api-service', 'port': {'number': 9100}}}}, {'path': '/testpayjp', 'pathType': 'Exact', 'backend': {'service': {'name': 'test-api-service', 'port': {'number': 9100}}}}, {'path': '/testworkplace', 'pathType': 'Exact', 'backend': {'service': {'name': 'test-api-service', 'port': {'number': 9100}}}}, {'path': '/api', 'pathType': 'Exact', 'backend': {'service': {'name': 'prod-api-service', 'port': {'number': 9000}}}}, {'path': '/payjp', 'pathType': 'Exact', 'backend': {'service': {'name': 'prod-api-service', 'port': {'number': 9000}}}}, {'path': '/workplace', 'pathType': 'Exact', 'backend': {'service': {'name': 'prod-api-service', 'port': {'number': 9000}}}}]}}, {'host': 'papi.toss-online.com', 'http': {'paths': [{'path': '/api', 'pathType': 'Exact', 'backend': {'service': {'name': 'prod-api-service', 'port': {'number': 9000}}}}]}}]}}

simple = {'apiVersion': 'networking.k8s.io/v1  the is the longest line of them all, twoa three foru five six seven eight nine ten ', 'kind': 'Ingress', 'metadata': {'name': 'toss-ingress', 'namespace': 'toss', 'annotations': {'kubernetes.io/ingress.global-static-ip-name': 'toss-ip', 'nginx.ingress.kubernetes.io/proxy-read-timeout': '120'}}, 'spec': {'tls': [{'secretName': 'toss-online-com-certbot'}], 'rules': [{'host': 'm-test.tos-land.net', 'http': {'paths': [{'path': '/', 'pathType': 'Prefix', 'backend': {'service': {'name': 'prod-api-service', 'port': {'number': 9000}}}}]}}, {'host': 'tos-land.net', 'http': {'paths': [{'path': '/', 'pathType': 'Prefix', 'backend': {'service': {'name': 'prod-api-service', 'port': {'number': 9000}}}}]}} ]}}

def make_profile(x):    
    print.pretty(x)

    

def show(f1, f2):

    p1 = pstats.Stats(f1)
    p1.strip_dirs()
    p2 = pstats.Stats(f2)
    p2.strip_dirs()
    
    t = Table(1,1,1,1,1,1,1,1)
    t.cell('c0',just='>') # Calls
    t.cell('c1',border='m:0', style='dem') # /Recursive
    t.cell('C5',just='>',style='1') #Filename
    t.cell('C6',border='m:0', style='dem') #lineno
    t.cell('C7',style='2') # Function
    t('Calls\t','\bdem /Recursive\t', 'Delta\t', 'TotTime\tCumTime\tFile\t','\bdem :lno','\tFunction\t')

    def _pct(chg):
        chg *= 100
        chgs = f'{chg:.1f}'
        return ('\br$' if chg < -1 else '\b;$' if chg < 3 else '\bg;$' if chg < 10 else '\bg!$', f'+{chgs}' if chg >= 0 else chgs) if chgs != '0.0' else (' ',)

    for k,v in sorted(p1.stats.items(), key=lambda x: x[1][3]):#x[1][3] / x[1][0]):
        v2 = p2.stats[k] if k in p2.stats else v
        t(v[0],'\t')
        if v[1]!=v[0]: t('/', v[1]) 
        t(*_pct((v2[1] - v[1])/v[1]), '\t')
        t(' ' , '\t')#t('\br$' if chg < -1 else '' if chg < 3 else '\by;$' if chg < 10 else '\bg!$', f'{chg:.1f}', '\t')
        #  time
        v2p = int(v[2]*1000//v[1])
        if v[2] >= 0.05:
            t(f'{v[2]:.1f}', f'\bdem  {v2p or ""}')
            t(*_pct((v2[2]-v[2])/v[2]), '\t')
        else:
            t(' \t')
        # Cumulative time 
        v3p = int(v[3]*1000//v[0])
        if v[3] >= 0.05:
            t(f'{v[3]:.1f}', f'\bdem  {v3p or ""}')
            t(*_pct((v2[3]-v[3])/v[3]), '\t')
        else:
            t(' \t')
        # Filename
        t(k[0], '\t', ':',k[1],'\t') if k[1] else t(' \t \t')
        # Function
        t(k[2],)
        for k in v[4]:
            t(' \b3$ ',k[0],':',k[1])
        t('\t')
        
    print(t)
    print(p1.total_tt, ' -> ', p2.total_tt,  '  ', *_pct((p2.total_tt-p1.total_tt)/p1.total_tt))




if __name__=='__main__':        
    if sys.argv[1] in 'hs':
        if len(sys.argv) > 2:
            cProfile.run('make_profile('+('hard' if sys.argv[1]=='h' else 'simple')+')', sys.argv[2])
        else:
            make_profile(hard if sys.argv[1]=='h' else simple)
        sys.exit(0)
    f1 = sys.argv[1]
    f2 = f1 if len(sys.argv) == 2 else sys.argv[2]
    show(f1,f2)



from .flex import Flex
from .hr import HR
from .borders import Bdr
from .context import Context



class Card(Flex, border='-', border_style='dem'):
    ''' Show a message in a card-like box.

    Use a tab (\\t) to separate the title from the body.
    
    >>> from print_ext.printer import Printer
    >>> Printer().card('\\tHello\\nWorld!')
    ┌────────┐
    │ Hello  │
    │ World! │
    └────────┘
    >>> card = Card('\berr$', 'Danger', '\b$ !\\t', "Don't hold plutonium\\nwith bare hands.")
    >>> Printer(lang='en')(card) 
    ┌┤ Danger! ├───────────┐
    │ Don't hold plutonium │
    │ with bare hands.     │
    └──────────────────────┘
    '''
    
    def __init__(self, *args, **kwargs):
        super().__init__(**kwargs)
        self.title = HR(parent=self, just='_', width_max=0) # FIXME: We need a way to stop propagation
        self._filling_title = True
        self.body = Bdr(border=(' ','m:01'))
        self(*args)


    def _process_rich_stream(self):
        while self.rich_stream:
            elctx = self.rich_stream.pop(0)
            if isinstance(elctx[0], Context): elctx[0].ctx_parent(None) # reparent to title or body
            if self._filling_title and elctx[0] == '\t':
                self._filling_title = False
                continue
            (self.title if self._filling_title else self.body).rich_stream.append(elctx)
            
    
    def calc_width(self):
        self._process_rich_stream()
        return max(self.title.width, self.body.width+2)


    def calc_height(self):
        self._process_rich_stream()
        return self.title.height + self.body.height + 1


    def flatten(self, w=0, h=0, **kwargs):
        self._process_rich_stream()
        w = w or self.width
        bdr = self['border']
        t = bdr.c[0]+bdr.t[1:3]+bdr.c[1]+bdr.t[4]+  bdr.c[4]+bdr.t[6:8]+bdr.c[5]+bdr.t[9]
        self.title['border'] = Bdr.dfn(bdr, t=t)
        title = list(self.title.flatten(w=w,h=0,**kwargs))
        h = h or self['height_max'] or 0
        if h and h > 2 and len(title) + 2 > h: title = list(self.title.flatten(w=w,h=1,**kwargs))
        if h: h -= len(title)
        yield from title
        body = Bdr(self.body, parent=self, border=('m:0111'))
        yield from body.flatten(w=w, h=h, **kwargs)
        


import re
from functools import reduce

class SGR():
    ''' Set Graphics Rendition

    String codes:
      ``0`` reset
      ``krgybmcw``  foreground color (blacK, Red, Green, Yellow, Blue, Magenta, Cyan, White)
      ``^krgybmcw`` Background color
      ``!`` / ``.`` Bold / Normal
      ``;`` / ``,`` Dim / Normal
      ``_`` / ``~`` underscore / No underscore
    '''
    __props__ = ('flags',)

    flags = ['fg','bg','bold','dim','ul']

    RE = re.compile(r'^(0)|(([krgybmcw]?)(\^([krgybmcw]))?([,;.!~_]*))$')

    def __init__(self, sval='', flags=None):
        if flags != None:
            self.flags = flags
            return
        # Init from a string
        m = SGR.RE.match(sval)
        if not m: raise ValueError(f"Invalid SGR value {sval!r}")
        self.flags = {}
        for v in m[6] or []:
            for k, g in {'dim':',;', 'ul':'~_', 'bold':'!.'}.items():
                if v in g:
                    if k in self.flags:
                        raise ValueError(f"Invalid SGR value {sval!r}. Can't set flag '{v}' because it is already set to '{self.flags[k]}'.")
                    self.flags[k] = v
        if m[3]: self.flags['fg'] = 30 + 'krgybmcw'.index(m[3])
        if m[4]: self.flags['bg'] = 40 + 'krgybmcw'.index(m[5])


    def __add__(self, other):
        if not (self and other): return other
        flags = dict(self.flags)
        flags.update(other.flags)
        return SGR(flags=flags)
    

    def __bool__(self):
        return bool(self.flags)
        

    def __getitem__(self, k):
        return self.flags.get(k,'')


    def diff(self, other=None):
        ''' other -> self '''
        other = other or SGR()
        codes = []
        if not self: return '\033[0m' if other else ''
        reset = (self['fg'] == '' and other['fg']) or (self['bg'] == '' and other['bg']) or (other['ul']=='_' and other['ul'] != self['ul']) or (other['dim']==';' and other['dim'] != self['dim']) or (other['bold']=='!' and other['bold'] != self['bold'])
        if self['bold'] == '!' and (reset or self['bold'] != other['bold']): codes.append(1)
        if self['dim'] == ';' and (reset or self['dim'] != other['dim']): codes.append(2)
        if self['ul'] == '_' and (reset or self['ul'] != other['ul']): codes.append(4)
        if self['fg'] and (reset or self['fg'] != other['fg']): codes.append(self['fg'])
        if self['bg'] and (reset or self['bg'] != other['bg']): codes.append(self['bg'])
        if reset: codes.insert(0,0)
        return '\033[' + ';'.join(map(str,codes)) + 'm'


    def __hash__(self):
        return reduce(lambda h, f: h ^ hash(self.flags.get(f,'')), SGR.flags)


    def __eq__(self, other):
        return all(other.flags.get(f,'') == self.flags.get(f,'') for f in SGR.flags)


    def __str__(self):
        s = ''
        if not self: return '0'
        if 'fg' in self.flags: s += 'krgybmcw'[self.flags['fg']-30]
        if 'bg' in self.flags: s += '^'+'krgybmcw'[self.flags['bg']-40]
        return s + ''.join(self.flags.get(k,'') for k in SGR.flags[2:])


    def __repr__(self):
        return f"SGR('{self}')"



import asyncio, pytest
from print_ext import print
from .testutil import printer


def test_progress_spinner_gone():
    o,p = printer()
    bar = p.progress()
    bar('\b1 Ready')
    bar('\b1 Set')
    bar('\b1 Go', done=True)
    print(o.getvalue())
    assert(o.getvalue() == 'Ready\nSet   \nGo  \n')
    


def test_spinner():
    print['ascii'] = False
    p = print.progress(steps=30)
    for i in range(30):
        p(f'\b{i%3+1} カタカナん {i}')
    p('done??', style='err', done=True)
    print('---')


async def many_bars():
    pass
    #go to the (start of the) previous line: \033[F
    #move along ncols: \03[{ncols}G


if __name__ == '__main__':
    asyncio.run(many_bars())



import pytest
from print_ext import pretty, PrettyException, Table, HR
from .testutil import printer, debug_dump


class Beautiful():
    def __init__(self, **kwargs):
        self.vals = kwargs
    
    def pretty(self, **kwargs):
        t = Table(0,0)
        self.vals.update(kwargs)
        for k,v in self.vals.items():
            t(k,'\t',pretty(v),'\t')
        return t 


@pytest.mark.skip()
def test_pretty_list():
    o,p = printer(color=True)
    l = pretty(['the',33,(99,1),{'a':39,'tens':[10,20,30,40]}])
    p(l)
    print(o.getvalue())
    assert(o.getvalue() == '')


def test_pretty_beautiful():
    o, p = printer()
    b = Beautiful(a=3, b=4)
    p.pretty(b, b=5, c=6)
    print(o.getvalue())
    assert(o.getvalue() == 'a 3\nb 5\nc 6\n')


def test_pretty_exception():
    o,p = printer()
    try:
        raise PrettyException(a=[1,2,3], b={'x':'hi','Hello': 'World'})
    except PrettyException as e:
        p(pretty(e))
        print(o.getvalue())
        assert(str(e) == 'PrettyException')
        assert(repr(e) == "PrettyException(a=[1, 2, 3], b={'x': 'hi', 'Hello': 'World'})")
        assert(o.getvalue() == 'PrettyException\na [0] 1\n  [1] 2\n  [2] 3\nb     x hi\n  Hello World\n')
    

def test_pretty_exception_no_pretty():
    class MyException(PrettyException):
        def pretty(self):
            return ''
    o,p = printer()
    try:
        raise MyException(a=[1,2,3], b={'x':'hi','Hello': 'World'})
    except MyException as e:
        p(pretty(e))
        print(o.getvalue())
        assert(str(e) == "MyException(a=[1, 2, 3], b={'x': 'hi', 'Hello': 'World'})")
        assert(repr(e) == "MyException(a=[1, 2, 3], b={'x': 'hi', 'Hello': 'World'})")
        assert(o.getvalue() == '')


def test_pretty_exception_no_style():
    class MyException(PrettyException):
        def pretty(self):
            return HR('\b1 hello')
    o,p = printer(ascii=False, color=True)
    try:
        raise MyException()
    except MyException as e:
        p(pretty(e))
        assert(str(e) == "-[ hello ]-")        
        assert(repr(e) == "MyException()")
        assert(o.getvalue() == '─┤ \x1b[33mhello\x1b[0m ├─\n')
